using System;
using System.Collections.Generic;

using Microsoft.Xna.Framework;

using Pong;

using Syntriax.Engine.Core;
using Syntriax.Engine.Core.Abstract;
using Syntriax.Engine.Graphics.TwoDimensional;
using Syntriax.Engine.Physics2D.Abstract;

namespace Syntriax.Engine.Physics2D;

public class PhysicsEngine2D : IPhysicsEngine2D
{
    private List<IRigidBody2D> rigidBodies = new List<IRigidBody2D>(32);
    private List<ICollider2D> colliders = new List<ICollider2D>(64);

    private int _iterationCount = 1;


    public int IterationCount { get => _iterationCount; set => _iterationCount = value < 1 ? 1 : value; }

    public void AddRigidBody(IRigidBody2D rigidBody)
    {
        if (rigidBodies.Contains(rigidBody))
            return;

        rigidBodies.Add(rigidBody);

        foreach (var collider2D in rigidBody.BehaviourController.GetBehaviours<ICollider2D>())
            colliders.Add(collider2D);

        rigidBody.BehaviourController.OnBehaviourAdded += OnBehaviourAdded;
        rigidBody.BehaviourController.OnBehaviourRemoved += OnBehaviourRemoved;
    }

    public void RemoveRigidBody(IRigidBody2D rigidBody)
    {
        rigidBodies.Remove(rigidBody);
    }

    public void Step(float deltaTime)
    {
        float intervalDeltaTime = deltaTime / IterationCount;

        for (int iterationIndex = 0; iterationIndex < IterationCount; iterationIndex++)
        {
            foreach (var rigidBody in rigidBodies)
                StepRigidBody(rigidBody, intervalDeltaTime);


            foreach (var collider in colliders)
                collider.Recalculate();

            for (int i = colliders.Count - 1; i >= 0; i--)
                CheckCollisions(colliders[i], colliders, (c1, c2) =>
                {
                    if (c1.RigidBody2D is IRigidBody2D c1RigidBody) c1RigidBody.Velocity = -c1RigidBody.Velocity;
                    if (c2.RigidBody2D is IRigidBody2D c2RigidBody) c2RigidBody.Velocity = -c2RigidBody.Velocity;
                });
        }
    }

    private void CheckCollisions(ICollider2D collider2D, List<ICollider2D> collider2Ds, Action<ICollider2D, ICollider2D> OnCollisionDetectedAction)
    {
        for (int i = collider2Ds.Count - 1; i >= 0; i--)
        {
            ICollider2D collider2DItem = collider2Ds[i];
            if (collider2DItem == collider2D)
                continue;

            foreach (var vertex in collider2DItem.Vertices)
                if (collider2D.CheckCollision(vertex))
                    OnCollisionDetectedAction?.Invoke(collider2D, collider2DItem);
        }
    }

    private void StepRigidBody(IRigidBody2D rigidBody, float intervalDeltaTime)
    {
        Vector2 nextPosition = rigidBody.Transform.Position;
        nextPosition += rigidBody.Velocity * intervalDeltaTime;

        rigidBody.Transform.Position = nextPosition;
    }

    private void OnBehaviourAdded(IBehaviourController controller, IBehaviour behaviour)
    {
        if (behaviour is not ICollider2D collider2D)
            return;

        colliders.Add(collider2D);
    }

    private void OnBehaviourRemoved(IBehaviourController controller, IBehaviour behaviour)
    {
        if (behaviour is not ICollider2D collider2D)
            return;

        colliders.Remove(collider2D);
    }
}
